/* C implementation for binary trees, originally implemented by Michael Juenger (Cologne). 
   Minor changes by Frauke Liers (Cologne) 08/2011*/

#include <stdio.h>
#include <malloc.h>

typedef int infotype;

typedef struct node_struct {
  int key;
  int level;
  struct node_struct *father;
  struct node_struct *leftson;
  struct node_struct *rightson;
} node;

void initialize();
void insert();
void delete();
void preorder();
void deletetree();
void visit();
node * recsearch();
node * search();
node * minimum();
node * maximum();
node * successor();


int main()

{

  node * root;
  node * newnode;
  int i,n;
  int key;

  scanf("%d",&n);

  initialize(&root);

  for(i = 1; i <= n; i++){
    scanf("%d",&key);
    newnode = (node *) malloc(sizeof(node)); newnode->key = key; insert(&root,newnode);
  }

  printf("maximum key is %d\n", maximum(root)->key);  

  printf("successor of root with key %d is element with key %d\n", root->key, successor(root)->key);  

  newnode = search(root,10);
  if(newnode==NULL) {
    printf("searching for element with key 10 was without success!\n");
  }
  else{
    printf("successor of element with key %d is element with key %d\n", newnode->key, successor(newnode)->key);  
  }

  printf("preorder\n"); preorder(root);
  
  deletetree(root);
  return 0;
}

void initialize(node **root)
{
  *root = NULL;
}

void insert(node **root, node *q)
/* insert *q in tree with root **root */
{
  node *p, *r;
  r = NULL;
  p = *root;
  while (p!=NULL) {
    r = p;
    if (q->key<p->key) p = p->leftson; else p = p->rightson;
  }
  q->father = r;
  q->leftson = NULL;
  q->rightson = NULL;
  if (r==NULL) {
    q->level = 0;
    *root = q;
  }
  else {
    q->level = r->level+1;
    if (q->key<r->key) r->leftson = q; 
    else r->rightson = q;
  }
}

node * recsearch(node *p, int k)
{
  if ( (p==NULL) || (p->key==k) ) return p;
  else if (k<p->key) return search(p->leftson,k); else return search(p->rightson,k);
}

node * search(node *p, int k)
{
  while ( (p!=NULL) && (p->key!=k) ) 
    if (k<p->key) p = p->leftson; else p = p->rightson;
  return p;
}

void delete(node **root, node *q)
/* delete *q from tree with root **root */
{
  node *p, *r;
  /* determine a node r to delete */
  if ( (q->leftson==NULL) || (q->rightson==NULL) ) r = q; else r = successor(q);
  /* let p point to the child of r. p=NULL, if r does not have children */
  if (r->leftson!=NULL) p = r->leftson; else p = r->rightson;
  /* cut out *r */
  if (p!=NULL) p->father = r->father;
  if (r->father==NULL) *root = p;
  else if (r==r->father->leftson) r->father->leftson = p; else r->father->rightson = p;
  /* copy data from *r to *q, if necessary */
  if (r!=q) {
    q->key = r->key;
  }
  free(r);
}

void preorder(node *p)
{
  if (p!=NULL) {
    visit(p);
    preorder(p->leftson);
    preorder(p->rightson);
  }
}

void deletetree(node *p)
{
  if (p!=NULL) {
    deletetree(p->leftson);
    deletetree(p->rightson);
    free(p);
    p=NULL;
  }
}


void visit(node *p)
{
  printf("Visiting node with key %d and level %d\n",p->key, p->level);
}

node * minimum(node *p)
{
  while (p->leftson!=NULL) p = p->leftson;
  return p;
}

node * maximum(node *p)
{
  while (p->rightson!=NULL) p = p->rightson;
  return p;
}


/* the successor function takes two paths:

   1. If the node has right child then return the left most node in the right sub tree.
   2. Else, it goes up from node until we find a node that is to the left of its parent. */

node * successor(node *p)
{
  node *q;
  /* return the left most node in the right sub tree */
  if (p->rightson!=NULL) return minimum(p->rightson);
  else {  /* go up from node until we find a node that is the left of its parent */
    q = p->father;
    while ( (q!=NULL) && (p==q->rightson) ) {
      p = q;
      q = q->father;
    }
    return q;
  }
}




