/* implementation for Labyrinth exercise, written by Frauke Liers (Cologne) in 08/2011. */
/* for ease of presentation, all functions are in one file. For a larger project, 
   one puts, e.g. the stack implementation in a separate file */


#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>

//Coordinates of the fields are stored in a struct Coord with x- and y- coordinates. 

typedef struct {
    int x;
    int y;
} Coord;

// here comes an implementation of a stack:

typedef struct {
  Coord *stack;
  int stackpointer; 
  int maxinstack;
} Stack;


void initialize();
int stackempty();
int stackfull();
void push();
Coord pop();
Coord top();

int main(){

  //input (quadratic) labyrinth
    int max; //size of labyrinth
    scanf("%d",&max);
    int i,j;
    int L[max+2][max+2];

  //put 1s around the labyrinth
  for(i = 0; i < max+2; i++){
      L[0][i] = 1;
      L[max+1][i] = 1;
      L[i][0] =  1;
      L[i][max+1] = 1;
  }
  //input of labyrinth. 0 is a white field, 1 a black one (a wall)
  for (i =0; i < max; i++) {
      for (j = 0; j < max; j++) {
	scanf("%d",&L[i+1][j+1]);
      }
  }


  Coord sp; //start
  Coord zp; //end
  scanf("%d", &sp.x);
  scanf("%d", &sp.y);
  scanf("%d", &zp.x);
  scanf("%d", &zp.y);

  sp.x++; sp.y++;
  zp.x++; zp.y++;

  // output labyrinth
  printf("labyrinth\n");
  for ( i =0; i <= max+1; i++) {
      for ( j = 0; j <= max+1; j++) {
	printf("%d ",L[i][j]);
      }
      printf("\n");
  }
  printf("\n");

  printf("start %d, %d\n", sp.x,  sp.y);
  printf("end %d, %d\n", zp.x, zp.y);
  

  int found = 0; //path not yet found
   
  // initialize empty stack. We only push white fields into stack. Thus, the size of the stack is limited 
  // by max*max because this is the size of the labyrinth. 
  
  Stack *S = malloc(sizeof(Stack));
  initialize(S,(max)*(max));  

  push(S,sp); 
  //push start sp into the stack because we start searching from here

  while(!stackempty(S)) { 
    
      //check whether the first element in the stack is the end point. If yes, stop. 

      Coord position = top(S);
      if((position.x == zp.x) && position.y == zp.y){
	  printf("Found Path!!!\n");
	  found = 1;
	  break;
      }
      
      //Otherwise: 
      
      //We are at point position with x-coordinate position.x and
      //y-coordinate position.y. We need to decide in which direction we walk such 
      // that in the end we have tried all possibilities up/down/left/right. 
      //We enumerate these directions by numbers

      // 1: down
      // 2: left
      // 3: up
      // 4: right

      //We increase the entry in field L[position.x][position.y] by one unit. 
      //Then we check the number in this field. This number shows in which direction
      // we will go next. We do this with case statements. If there is a 1 in the field, 
      // we to down etc. As we increase the value of the field entry in each iteration, 
      // we know that we have tried all directions once we encounter a value of 4 in the field. 
      // Then we can safely remove this field from the stack. 

      //What about the walls? Is it true that we reach all possible fields and that we 
      // only move along white fields? Yes. After the case statements, we check whether the 
      // NEW field is a white one. If yes, it is pushed into the stack. However, fields with entries 1 
      // are never pushed and considered as possible ways through the labyrinth. This means we 
      // always walk along white fields and not along walls. Once a new field is in the stack, 
      // the entry 0 does not matter any more because we know that it is a feasible field whose field 
      // entry we can increase by one if we stand on it. 

      L[position.x][position.y]++;
      switch (L[position.x][position.y]){
      case 1: position.x++; break;
      case 2: position.x--; break;
      case 3: position.y--; break;
      case 4: position.y++; break;
      case 5: pop(S); continue;
      }
      if (L[position.x][position.y] == 0)  push(S,position);
    
    }

  //If we have found a path, it is stored in reverse orientation in the stack. 

  if(found){
    printf("Found\n");
      while (!stackempty(S)){
	  Coord position = pop(S);
	  printf("( %d, %d )\n", position.x ,position.y);
	  
      }
  }
  free(S->stack);
  free(S);
  return 0;
}

void initialize(Stack *S, int max)
{
    
    S->maxinstack = max;
    S->stack = (Coord *) malloc(max*sizeof(Coord));
    S->stackpointer = 0;
}


int stackempty(Stack* S)
{
  if(S->stackpointer<=0) return 1;
  return 0;
}

int stackfull(Stack* S)
{
  if(S->stackpointer>=S->maxinstack) return 1;
  return 0;
}

void push(Stack * S,Coord v)
{
  if (stackfull(S)) printf("!!! Stack is full !!!\n");
    else {
	S->stack[S->stackpointer] = v;
	S->stackpointer++;
    }
}

Coord pop(Stack * S)
{
  if (stackempty(S)) {
    printf("!!! Stack is empty !!!\n");
    exit(0);
  }
  S->stackpointer--;
  return S->stack[S->stackpointer];
}

Coord top(Stack * S)
{
    if (stackempty(S)) {
      printf("!!! Stack is empty !!!\n");
      exit(0);
    }
    return S->stack[S->stackpointer-1];
}
